# This file is part of TALER
# (C) 2019 GNUnet e.V.
#
# Authors:
# Author: ng0 <ng0@taler.net>
# Author: Florian Dold <dold@taler.net>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE
# LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
# THIS SOFTWARE.
#
# SPDX-License-Identifier: 0BSD

from abc import ABC
import argparse
import os
import sys
import shlex
import logging
from distutils.spawn import find_executable
import subprocess
from dataclasses import dataclass

"""
This module aims to replicate a small GNU Coding Standards
configure script, taylored at projects in GNU Taler. We hope it
can be of use outside of GNU Taler, hence it is dedicated to the
public domain ('0BSD').
It takes a couple of arguments on the commandline equivalent to
configure by autotools, in addition some environment variables
xan take precedence over the switches. In the absence of switches,
/usr/local is assumed as the PREFIX.
When  all data from tests are gathered, it generates a config.mk
Makefile fragement, which is the processed by a Makefile (usually) in
GNU Make format.
"""


# TODO: We need a smallest version argument.

class Tool(ABC):
    def args(self):
        ...

    def check(self, buildconfig):
        ...


class BuildConfig:
    def __init__(self):
        # Pairs of (key, value) for config.mk variables
        self.make_variables = []
        self.tools = []
        self.tool_results = {}
        self.args = None
        self.prefix_enabled = False
        self.configmk_enabled = False

    def add_tool(self, tool):
        if isinstance(tool, Tool):
            self.tools.append(tool)
        else:
            raise Exception("Not a tool instance: " + repr(tool))

    def _set_tool(self, name, value, version=None):
        self.tool_results[name] = (value, version)

    def enable_prefix(self):
        """If enabled, process the --prefix argument."""
        self.prefix_enabled = True

    def enable_configmk(self):
        """If enabled, output the config.mk makefile fragment."""
        self.configmk_enabled = True

    def run(self):
        parser = argparse.ArgumentParser()
        if self.prefix_enabled:
            parser.add_argument(
                "--prefix",
                type=str,
                default="/usr/local",
                help="Directory prefix for installation",
            )
        for tool in self.tools:
            tool.args(parser)
        args = self.args = parser.parse_args()

        for tool in self.tools:
            res = tool.check(self)
            if not res:
                print(f"Error: tool {tool.name} not available")
                if hasattr(tool, "hint"):
                    print(f"Hint: {tool.hint}")

        for tool in self.tools:
            path, version = self.tool_results[tool.name]
            if version is None:
                print(f"found {tool.name} as {path}")
            else:
                print(f"found {tool.name} as {path} (version {version})")

        if self.configmk_enabled:
            with open("config.mk", "w") as f:
                print("writing config.mk")
                f.write("# this makefile fragment is autogenerated by configure.py\n")
                if self.prefix_enabled:
                    f.write(f"prefix = {args.prefix}\n")
                for tool in self.tools:
                    path, version = self.tool_results[tool.name]
                    f.write(f"{tool.name} = {path}\n")


def existence(name):
    return find_executable(name) is not None


class YarnTool(Tool):
    name = "yarn"
    description = "The yarn package manager for node"

    def args(self, parser):
        parser.add_argument("--with-yarn", action="store")

    def check(self, buildconfig):
        yarn_arg = buildconfig.args.with_yarn
        if yarn_arg is not None:
            buildconfig._set_tool("yarn", yarn_arg)
            return True
        if existence("yarn"):
            p1 = subprocess.run(
                ["yarn", "help"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE
            )
            if "No such file or directory" in p1.stdout.decode("utf-8"):
                if existence("cmdtest"):
                    buildconfig._warn(
                        "cmdtest is installed, this can lead to known issues with yarn."
                    )
                buildconfig._error(
                    "You seem to have the wrong kind of 'yarn' installed.\n"
                    "Please remove the conflicting binary before proceeding"
                )
                return False
            yarn_version = tool_version("yarn --version")
            buildconfig._set_tool("yarn", "yarn", yarn_version)
            return True
        elif existence("yarnpkg"):
            yarn_version = tool_version("yarnpkg --version")
            buildconfig._set_tool("yarn", "yarnpkg", yarn_version)
            return True
        return False


def tool_version(name):
    return subprocess.getstatusoutput(name)[1]


class EmscriptenTool:
    def args(self, parser):
        pass

    def check(self, buildconfig):
        if existence("emcc"):
            emscripten_version = tool_version("emcc --version")
            buildconfig._set_tool("emcc", "emcc", emscripten_version)
            return True
        return False


class PyBabelTool(Tool):
    name = "pybabel"

    def args(self, parser):
        parser.add_argument(
            "--with-pybabel", type=str, help="name of the pybabel executable"
        )

    def check(self, buildconfig):
        # No suffix. Would probably be cheaper to do this in
        # the dict as well. We also need to check the python
        # version it was build against (TODO).
        if existence("pybabel"):
            import babel
            pybabel_version = babel.__version__
            buildconfig._set_tool("pybabel", "pybabel", pybabel_version)
            return True
        else:
            # Has suffix, try suffix. We know the names in advance,
            # so use a dictionary and iterate over it. Use enough names
            # to safe updating this for another couple of years.
            #
            # Food for thought: If we only accept python 3.7 or higher,
            # is checking pybabel + pybabel-3.[0-9]* too much and could
            # be broken down to pybabel + pybabel-3.7 and later names?
            version_dict = {
                "3.0": "pybabel-3.0",
                "3.1": "pybabel-3.1",
                "3.2": "pybabel-3.2",
                "3.3": "pybabel-3.3",
                "3.4": "pybabel-3.4",
                "3.5": "pybabel-3.5",
                "3.6": "pybabel-3.6",
                "3.7": "pybabel-3.7",
                "3.8": "pybabel-3.8",
                "3.9": "pybabel-3.9",
                "4.0": "pybabel-4.0",
            }
            for key, value in version_dict.items():
                if existence(value):
                    # FIXME: This version reporting is slightly off
                    # FIXME: and only maps to the suffix.
                    pybabel_version = key
                    buildconfig._set_tool("pybabel", value, pybabel_version)
                    return True


class PythonTool(Tool):
    # This exists in addition to the files in sh, so that
    # the Makefiles can use this value instead.
    name = "python"

    def args(self, parser):
        parser.add_argument(
            "--with-python", type=str, help="name of the python executable"
        )

    def check(self, buildconfig):
        # No suffix. Would probably be cheaper to do this in
        # the dict as well. We need at least version 3.7.
        if existence("python") and (shlex.split(subprocess.getstatusoutput("python --version")[1])[1] >= '3.7'):
            # python might not be python3. It might not even be
            # python 3.x.
            python_version = shlex.split(subprocess.getstatusoutput("python --version")[1])[1]
            if python_version >= '3.7':
                buildconfig._set_tool("python", "python", python_version)
                return True
        else:
            # Has suffix, try suffix. We know the names in advance,
            # so use a dictionary and iterate over it. Use enough names
            # to safe updating this for another couple of years.
            #
            # Food for thought: If we only accept python 3.7 or higher,
            # is checking pybabel + pybabel-3.[0-9]* too much and could
            # be broken down to pybabel + pybabel-3.7 and later names?
            version_dict = {
                "3.7": "python3.7",
                "3.8": "python3.8",
                "3.9": "python3.9",
                "4.0": "python4.0",
            }
            for key, value in version_dict.items():
                if existence(value):
                    python3_version = key
                    buildconfig._set_tool("python", value, python3_version)
                    return True


# TODO: Make this really optional, not use a hack ("true").
class BrowserTool(Tool):
    name = "browser"

    def args(self, parser):
        parser.add_argument(
            "--with-browser", type=str, help="name of your webbrowser executable"
        )

    def check(self, buildconfig):
        browser_dict = {
            "ice": "icecat",
            "ff": "firefox",
            "chg": "chrome",
            "ch": "chromium",
            "o": "opera",
            "t": "true"
        }
        if "BROWSER" in os.environ:
            buildconfig._set_tool("browser", os.environ["BROWSER"])
            return True
        for value in browser_dict.values():
            if existence(value):
                buildconfig._set_tool("browser", value)
                return True


class NodeJsTool(Tool):
    name = "node"
    hint = "If you are using Ubuntu Linux or Debian Linux, try installing the\nnode-legacy package or symlink node to nodejs."

    def args(self, parser):
        pass

    def check(self, buildconfig):
        if existence("node") is None:
            return False
        if (
            subprocess.getstatusoutput(
                "node -p 'process.exit(!(/v([0-9]+)/.exec(process.version)[1] >= 4))'"
            )[1]
            != ""
        ):
            buildconfig._warn("your node version is too old, use Node 4.x or newer")
            return False
        node_version = tool_version("node --version")
        buildconfig._set_tool("node", "node", version=node_version)
        return True


class PosixTool(Tool):
    def __init__(self, name):
        self.name = name

    def args(self, parser):
        pass

    def check(self, buildconfig):
        found = existence("find")
        if found:
            buildconfig._set_tool(self.name, self.name)
            return True
        return False



#!/usr/bin/python3
#    This file is part of GNUnet.
#    (C) 2011, 2018 Christian Grothoff (and other contributing authors)
#
#    GNUnet is free software: you can redistribute it and/or modify it
#    under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNUnet is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    SPDX-License-Identifier: AGPL3.0-or-later
#
# Finds any gnunet processes still running in the system and kills them
#
# gnunet janitor can be used by invoking `make' like this:
# TESTS_ENVIRONMENT='${top_srcdir}/contrib/scripts/gnunet_janitor.py &&' make check

import os
import re
import subprocess
import sys
import shutil
import time
import signal
import terminate


def get_process_list():
    result = []
    pids = [pid for pid in os.listdir('/proc') if pid.isdigit()]
    for pid in pids:
        with open(os.path.join('/proc', pid, 'cmdline'), 'rb') as p:
            cmdline = p.read().split('\x00')
            if len(cmdline) > 0:
                result.append((pid, cmdline[0]))
    return result


def main():
    procs = get_process_list()
    gnunet_procs = []
    for p in procs:
        if re.match(r'gnunet-.+', p[1]):
            gnunet_procs.append(p)
    for p in gnunet_procs:
        if re.match(r'gnunet-service-arm', p[1]):
            print("killing arm process {0:5} {1}".format(p[0], p[1]))
            try:
                terminate.safe_terminate_process_by_pid(int(p[0]), 1)
            except OSError as e:
                print("failed: {0}".format(e))
                pass
    for p in gnunet_procs:
        if not re.match(r'gnunet-service-arm', p[1]):
            print("killing non-arm process {0:5} {1}".format(p[0], p[1]))
            try:
                terminate.safe_terminate_process_by_pid(int(p[0]), 1)
            except OSError as e:
                print("failed: {0}".format(e))
                pass


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
#
# This testcase simply checks that the DHT command-line tools work.
# It launches a single peer, stores a value "testdata" under "testkey",
# and then gives the system 50 ms to fetch it.
#
# This could fail if
# - command line tool interfaces fail
# - DHT plugins for storage are not installed / working
# - block plugins for verification (the test plugin) is not installed
#
# The code does NOT depend on DHT routing or any actual P2P functionality.
#

import os
import sys
import shutil
import re
import subprocess
import time
import tempfile

os.environ["PATH"] = "/home/schanzen/gnunet/build/bin" + ":" + os.environ["PATH"]

if os.name == "nt":
    tmp = os.getenv("TEMP")
else:
    tmp = "/tmp"

if os.name == 'nt':
    get = './gnunet-dht-get.exe'
    put = './gnunet-dht-put.exe'
    arm = 'gnunet-arm.exe'
else:
    get = './gnunet-dht-get'
    put = './gnunet-dht-put'
    arm = 'gnunet-arm'

cfgfile = 'test_dht_api_peer1.conf'
run_get = [get, '-c', cfgfile]
run_put = [put, '-c', cfgfile]
run_arm = [arm, '-c', cfgfile]
debug = os.getenv('DEBUG')
if debug:
    run_arm += [debug.split(' ')]


def cleanup(exitcode):
    sys.exit(exitcode)


def sub_run(args, want_stdo=True, want_stde=False, nofail=False):
    if want_stdo:
        stdo = subprocess.PIPE
    else:
        stdo = None
    if want_stde:
        stde = subprocess.PIPE
    else:
        stde = None
    p = subprocess.Popen(args, stdout=stdo, stderr=stde)
    stdo, stde = p.communicate()
    if not nofail:
        if p.returncode != 0:
            sys.exit(p.returncode)
    return (p.returncode, stdo, stde)


def fail(result):
    print(result)
    r_arm(['-e'], want_stdo=False)
    cleanup(1)


def r_something(to_run, extra_args, failer=None, normal=True, **kw):
    rc, stdo, stde = sub_run(to_run + extra_args, nofail=True, **kw)
    if failer is not None:
        failer(to_run + extra_args, rc, stdo, stde, normal)
    return (rc, stdo, stde)


def r_arm(extra_args, **kw):
    return r_something(run_arm, extra_args, **kw)


def r_get(extra_args, **kw):
    return r_something(run_get, extra_args, **kw)


def r_put(extra_args, **kw):
    return r_something(run_put, extra_args, **kw)


def end_arm_failer(command, rc, stdo, stde, normal):
    if normal:
        if rc != 0:
            fail(
                "FAIL: error running {}\nCommand output was:\n{}\n{}".format(
                    command, stdo, stde
                )
            )
    else:
        if rc == 0:
            fail(
                "FAIL: expected error while running {}\nCommand output was:\n{}\n{}"
                .format(command, stdo, stde)
            )


def print_only_failer(command, rc, stdo, stde, normal):
    if normal:
        if rc != 0:
            print(
                "FAIL: error running {}\nCommand output was:\n{}\n{}".format(
                    command, stdo, stde
                )
            )
            cleanup(1)
    else:
        if rc == 0:
            print(
                "FAIL: expected error while running {}\nCommand output was:\n{}\n{}"
                .format(command, stdo, stde)
            )
            cleanup(1)


print("TEST: Starting ARM...", end='')
r_arm(['-s'], failer=end_arm_failer, want_stdo=False, want_stde=False)
print("PASS")
time.sleep(1)

print("TEST: Testing put...", end='')
r_put(['-k', 'testkey', '-d', 'testdata', '-t', '8'], failer=end_arm_failer)
print("PASS")
time.sleep(1)

print("TEST: Testing get...", end='')
rc, stdo, stde = r_get(['-k', 'testkey', '-T', '50 ms', '-t', '8'],
                       want_stdo=True,
                       failer=end_arm_failer)
stdo = stdo.decode('utf-8').replace('\r', '').splitlines()
expect = "Result 0, type 8:\ntestdata".splitlines()
if len(stdo) != 2 or len(expect
                         ) != 2 or stdo[0] != expect[0] or stdo[1] != expect[1]:
    fail("output `{}' differs from expected `{}'".format(stdo, expect))
print("PASS")

r_arm(['-e', '-d'], failer=print_only_failer)



#!/usr/bin/python3

import os
import sys
import difflib
import filecmp


def getdiff(old, new):
    diff = []
    with open(old) as a:
        with open(new) as b:
            for l in difflib.unified_diff(
                a.read().splitlines(),
                b.read().splitlines()
            ):
                diff.append(l)
    return diff


def dc_getdiff(dc, old, new):
    diff = []
    for f in dc.left_only:
        diff.append("Only in {}: {}".format(old, f))
    for f in dc.right_only:
        diff.append("Only in {}: {}".format(new, f))
    for f in dc.diff_files:
        r = getdiff(os.path.join(old, f), os.path.join(new, f))
        diff.extend(r)
    for dn, dc in list(dc.subdirs.items()):
        r = dc_getdiff(dc, os.path.join(old, dn), os.path.join(new, dn))
        diff.extend(r)
    return diff


def dcdiff(old, new):
    dc = filecmp.dircmp(old, new)
    diff = dc_getdiff(dc, old, new)
    return diff


def main():
    for l in dcdiff(sys.argv[1], sys.argv[2]):
        print(l)


if __name__ == '__main__':
    main()



from talerbuildconfig import *

b = BuildConfig()
b.enable_prefix()
b.enable_configmk()
b.add_tool(YarnTool())
b.add_tool(BrowserTool())
b.add_tool(PyBabelTool())
b.add_tool(NodeJsTool())
b.add_tool(PythonTool())
b.add_tool(PosixTool("find"))
b.add_tool(PosixTool("xargs"))
b.add_tool(PosixTool("msgmerge"))
b.run()



#!/usr/bin/python3
#    This file is part of GNUnet.
#    (C) 2011, 2018 Christian Grothoff (and other contributing authors)
#
#    GNUnet is free software: you can redistribute it and/or modify it
#    under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GNUnet is distributed in the hope that it will be useful, but
#    WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    SPDX-License-Identifier: AGPL3.0-or-later
#
# Utility module that implements safe process termination for W32.
# For other platforms it's equivalent to Popen.kill ()
# Requires pywin32 on W32.

import sys
import subprocess
import os


class dummyobj(object):
    pass


def safe_terminate_process_by_pid(pid, code):
    # XXX (F821): Undefined name 'SIGKILL'
    return os.kill(int(pid), SIGKILL)


def safe_terminate_process(proc, code):
    return proc.kill()



#!/usr/bin/python3
# This file is part of GNUnet.
# (C) 2013, 2018 Christian Grothoff (and other contributing authors)
#
# GNUnet is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GNUnet is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: AGPL3.0-or-later
#
# File:    gnunet-chk.py
# Brief:   Computes GNUNET style Content Hash Key for a given file
# Author:  Sree Harsha Totakura

from hashlib import sha512
import logging
import os
import getopt
import sys
from Crypto.Cipher import AES
from functools import reduce

# Defaults
DBLOCK_SIZE = (32 * 1024)  # Data block size

# Pick a multiple of 2 here to achive 8-byte alignment!  We also
# probably want DBlocks to have (roughly) the same size as IBlocks.
# With SHA-512, the optimal value is 32768 byte / 128 byte = 256 (128
# byte = 2 * 512 bits).  DO NOT CHANGE!
CHK_PER_INODE = 256

CHK_HASH_SIZE = 64  # SHA-512 hash = 512 bits = 64 bytes

CHK_QUERY_SIZE = CHK_HASH_SIZE  # Again a SHA-512 hash

GNUNET_FS_URI_PREFIX = "gnunet://fs/"  # FS CHK URI prefix

GNUNET_FS_URI_CHK_INFIX = "chk/"  # FS CHK URI infix


def encode_data_to_string(data):
    """Returns an ASCII encoding of the given data block like
    GNUNET_STRINGS_data_to_string() function.

    data: A bytearray representing the block of data which has to be encoded
    """
    echart = "0123456789ABCDEFGHIJKLMNOPQRSTUV"
    assert (None != data)
    assert (bytearray == type(data))
    size = len(data)
    assert (0 != size)
    vbit = 0
    wpos = 0
    rpos = 0
    bits = 0
    out = ""
    while (rpos < size) or (vbit > 0):
        if (rpos < size) and (vbit < 5):
            bits = (bits << 8) | data[rpos]  # eat 8 more bits
            rpos += 1
            vbit += 8
        if (vbit < 5):
            bits <<= (5 - vbit)  # zero-padding
            assert (vbit == ((size * 8) % 5))
            vbit = 5
        out += echart[(bits >> (vbit - 5)) & 31]
        wpos += 1
        vbit -= 5
    assert (0 == vbit)
    return out


def sha512_hash(data):
    """ Returns the sha512 hash of the given data.

    data: string to hash
    """
    hash_obj = sha512()
    hash_obj.update(data)
    return hash_obj.digest()


class AESKey(object):
    """Class for AES Keys. Contains the main key and the initialization
    vector. """

    key = None  # The actual AES key
    iv = None  # The initialization vector
    cipher = None  # The cipher object
    KEY_SIZE = 32  # AES 256-bit key = 32 bytes
    IV_SIZE = AES.block_size  # Initialization vector size (= AES block size)

    def __init__(self, passphrase):
        """Creates a new AES key.

        passphrase: string containing the passphrase to get the AES key and
        initialization vector
        """
        passphrase = bytearray(passphrase)
        self.key = bytearray(self.KEY_SIZE)
        self.iv = bytearray(self.IV_SIZE)
        if (len(passphrase) > self.KEY_SIZE):
            self.key = passphrase[:self.KEY_SIZE]
            passphrase = passphrase[self.KEY_SIZE:]
            if (len(passphrase) > self.IV_SIZE):
                self.iv = passphrase[:self.IV_SIZE]
            else:
                self.iv[0:len(passphrase)] = passphrase
        else:
            self.key[0:len(passphrase)] = passphrase
        self.key = str(self.key)
        self.iv = str(self.iv)
        assert (len(self.key) == self.KEY_SIZE)
        assert (len(self.iv) == self.IV_SIZE)


def setup_aes_cipher_(aes_key):
    """Initializes the AES object with settings similar to those in GNUnet.

    aes_key: the AESKey object
    Returns the newly initialized AES object
    """
    return AES.new(aes_key.key, AES.MODE_CFB, aes_key.iv, segment_size=128)


def aes_pad_(data):
    """Adds padding to the data such that the size of the data is a multiple of
    16 bytes

    data: the data string
    Returns a tuple:(pad_len, data). pad_len denotes the number of bytes added
    as padding; data is the new data string with padded bytes at the end
    """
    pad_len = len(data) % 16
    if (0 != pad_len):
        pad_len = 16 - pad_len
        pad_bytes = bytearray(15)
        data += str(pad_bytes[:pad_len])
    return (pad_len, data)


def aes_encrypt(aes_key, data):
    """Encrypts the given data using AES.

    aes_key: the AESKey object to use for AES encryption
    data: the data string to encrypt
    """
    (pad_len, data) = aes_pad_(data)
    cipher = setup_aes_cipher_(aes_key)
    enc_data = cipher.encrypt(data)
    if (0 != pad_len):
        enc_data = enc_data[:-pad_len]
    return enc_data


def aes_decrypt(aes_key, data):
    """Decrypts the given data using AES

    aes_key: the AESKey object to use for AES decryption
    data: the data string to decrypt
    """
    (pad_len, data) = aes_pad_(data)
    cipher = setup_aes_cipher_(aes_key)
    ptext = cipher.decrypt(data)
    if (0 != pad_len):
        ptext = ptext[:-pad_len]
    return ptext


class Chk(object):
    """Class for the content hash key."""
    key = None
    query = None
    fsize = None

    def __init__(self, key, query):
        assert (len(key) == CHK_HASH_SIZE)
        assert (len(query) == CHK_QUERY_SIZE)
        self.key = key
        self.query = query

    def setSize(self, size):
        self.fsize = size

    def uri(self):
        sizestr = repr(self.fsize)
        if isinstance(self.fsize, int):
            sizestr = sizestr[:-1]
        return GNUNET_FS_URI_PREFIX + GNUNET_FS_URI_CHK_INFIX + \
            encode_data_to_string(bytearray(self.key)) + "." + \
            encode_data_to_string(bytearray(self.query)) + "." + \
            sizestr


def compute_depth_(size):
    """Computes the depth of the hash tree.

    size: the size of the file whose tree's depth has to be computed
    Returns the depth of the tree. Always > 0.
    """
    depth = 1
    fl = DBLOCK_SIZE
    while (fl < size):
        depth += 1
        if ((fl * CHK_PER_INODE) < fl):
            return depth
        fl = fl * CHK_PER_INODE
    return depth


def compute_tree_size_(depth):
    """Calculate how many bytes of payload a block tree of the given depth MAY
     correspond to at most (this function ignores the fact that some blocks will
     only be present partially due to the total file size cutting some blocks
     off at the end).

     depth: depth of the block.  depth==0 is a DBLOCK.
     Returns the number of bytes of payload a subtree of this depth may
     correspond to.
     """
    rsize = DBLOCK_SIZE
    for cnt in range(0, depth):
        rsize *= CHK_PER_INODE
    return rsize


def compute_chk_offset_(depth, end_offset):
    """Compute the offset of the CHK for the current block in the IBlock
    above

    depth: depth of the IBlock in the tree (aka overall number of tree levels
             minus depth); 0 == DBLOCK
    end_offset: current offset in the overall file, at the *beginning* of the
                  block for DBLOCK (depth == 0), otherwise at the *end* of the
                  block (exclusive)
    Returns the offset in the list of CHKs in the above IBlock
    """
    bds = compute_tree_size_(depth)
    if (depth > 0):
        end_offset -= 1
    ret = end_offset // bds
    return ret % CHK_PER_INODE


def compute_iblock_size_(depth, offset):
    """Compute the size of the current IBLOCK.  The encoder is triggering the
    calculation of the size of an IBLOCK at the *end* (hence end_offset) of its
    construction.  The IBLOCK maybe a full or a partial IBLOCK, and this
    function is to calculate how long it should be.

    depth: depth of the IBlock in the tree, 0 would be a DBLOCK, must be > 0
             (this function is for IBLOCKs only!)
    offset: current offset in the payload (!) of the overall file, must be > 0
              (since this function is called at the end of a block).
    Returns the number of elements to be in the corresponding IBlock
    """
    assert (depth > 0)
    assert (offset > 0)
    bds = compute_tree_size_(depth)
    mod = offset % bds
    if mod is 0:
        ret = CHK_PER_INODE
    else:
        bds /= CHK_PER_INODE
        ret = mod // bds
        if (mod % bds) is not 0:
            ret += 1
    return ret


def compute_rootchk(readin, size):
    """Returns the content hash key after generating the hash tree for the given
    input stream.

    readin: the stream where to read data from
    size: the size of data to be read
    """
    depth = compute_depth_(size)
    current_depth = 0
    chks = [None] * (depth * CHK_PER_INODE)  # list buffer
    read_offset = 0
    logging.debug("Begining to calculate tree hash with depth: " + repr(depth))
    while True:
        if (depth == current_depth):
            off = CHK_PER_INODE * (depth - 1)
            assert (chks[off] is not None)
            logging.debug("Encoding done, reading CHK `" + chks[off].query + \
                          "' from " + repr(off) + "\n")
            uri_chk = chks[off]
            assert (size == read_offset)
            uri_chk.setSize(size)
            return uri_chk
        if (0 == current_depth):
            pt_size = min(DBLOCK_SIZE, size - read_offset)
            try:
                pt_block = readin.read(pt_size)
            except IOError:
                logging.warning("Error reading input file stream")
                return None
        else:
            pt_elements = compute_iblock_size_(current_depth, read_offset)
            pt_block = ""
            pt_block = \
                reduce((lambda ba, chk:
                        ba + (chk.key + chk.query)),
                       chks[(current_depth - 1) * CHK_PER_INODE:][:pt_elements],
                       pt_block)
            pt_size = pt_elements * (CHK_HASH_SIZE + CHK_QUERY_SIZE)
        assert (len(pt_block) == pt_size)
        assert (pt_size <= DBLOCK_SIZE)
        off = compute_chk_offset_(current_depth, read_offset)
        logging.debug("Encoding data at offset " + repr(read_offset) + \
                      " and depth " + repr(current_depth) + " with block " \
                      "size " + repr(pt_size) + " and target CHK offset " + \
                      repr(current_depth * CHK_PER_INODE))
        pt_hash = sha512_hash(pt_block)
        pt_aes_key = AESKey(pt_hash)
        pt_enc = aes_encrypt(pt_aes_key, pt_block)
        pt_enc_hash = sha512_hash(pt_enc)
        chk = Chk(pt_hash, pt_enc_hash)
        chks[(current_depth * CHK_PER_INODE) + off] = chk
        if (0 == current_depth):
            read_offset += pt_size
            if (read_offset == size) or \
                    (0 == (read_offset % (CHK_PER_INODE * DBLOCK_SIZE))):
                current_depth += 1
        else:
            if (CHK_PER_INODE == off) or (read_offset == size):
                current_depth += 1
            else:
                current_depth = 0


def chkuri_from_path(path):
    """Returns the CHK URI of the file at the given path.

    path: the path of the file whose CHK has to be calculated
    """
    size = os.path.getsize(path)
    readin = open(path, "rb")
    chk = compute_rootchk(readin, size)
    readin.close()
    return chk.uri()


def usage():
    """Prints help about using this script."""
    print(
        """
Usage: gnunet-chk.py [options] file
Prints the Content Hash Key of given file in GNUNET-style URI.

Options:
    -h, --help                : prints this message
"""
    )


if '__main__' == __name__:
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h", ["help"])
    except getopt.GetoptError as err:
        print(err)
        print("Exception occured")
        usage()
        sys.exit(2)
    for option, value in opts:
        if option in ("-h", "--help"):
            usage()
            sys.exit(0)
    if len(args) != 1:
        print("Incorrect number of arguments passed")
        usage()
        sys.exit(1)
    print(chkuri_from_path(args[0]))
